@@ -1,6 +1,6 @@
 /*
  * tinylog
- * Copyright (C) 2018-2020 Nick Peng <pymumu@gmail.com>
+ * Copyright (C) 2018-2024 Nick Peng <pymumu@gmail.com>
  * https://github.com/pymumu/tinylog
  */
 #ifndef _GNU_SOURCE
@@ -19,6 +19,7 @@
 #include <string.h>
 #include <sys/resource.h>
 #include <sys/stat.h>
+#include <sys/syscall.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -36,7 +37,7 @@
 #define TLOG_TMP_LEN 128
 #define TLOG_LOG_SIZE (1024 * 1024 * 50)
 #define TLOG_LOG_COUNT 32
-#define TLOG_LOG_NAME_LEN 128
+#define TLOG_LOG_NAME_LEN 256
 #define TLOG_BUFF_LEN (PATH_MAX + TLOG_LOG_NAME_LEN * 3)
 #define TLOG_SUFFIX_GZ ".gz"
 #define TLOG_SUFFIX_LOG ""
@@ -45,6 +46,14 @@
 
 #define TLOG_SEGMENT_MAGIC 0xFF446154
 
+struct linux_dirent64 {
+    unsigned long long d_ino;
+    long long d_off;
+    unsigned short d_reclen;
+    unsigned char d_type;
+    char d_name[256];
+};
+
 struct tlog_log {
     char *buff;
     int buffsize;
@@ -70,8 +79,10 @@ struct tlog_log {
     int zip_pid;
     int multi_log;
     int logscreen;
+    int logscreen_color;
     int segment_log;
-    unsigned int max_line_size;
+    int max_line_size;
+    int print_errmsg;
 
     tlog_output_func output_func;
     void *private_data;
@@ -80,6 +91,7 @@ struct tlog_log {
     time_t last_waitpid;
     mode_t file_perm;
     mode_t archive_perm;
+    int mode_changed;
 
     int waiters;
     int is_exit;
@@ -99,6 +111,18 @@ struct tlog {
     tlog_log_output_func output_func;
     struct tlog_log *wait_on_log;
     int is_wait;
+    int output_no_prefix;
+    char gzip_cmd[PATH_MAX];
+
+    tlog_format_func root_format;
+
+    tlog_early_print_func tlog_early_print;
+    tlog_log_output_func early_print_output;
+    int early_print_disable;
+    int early_print_with_screen;
+    int early_print_no_prefix;
+    int early_print_color;
+    void *early_print_userptr;
 };
 
 struct tlog_segment_log_head {
@@ -133,10 +157,8 @@ typedef int (*list_callback)(const char *name, struct dirent *entry, void *user)
 typedef int (*vprint_callback)(char *buff, int maxlen, void *userptr, const char *format, va_list ap);
 
 static struct tlog tlog;
-static int tlog_disable_early_print = 0;
 static tlog_level tlog_set_level = TLOG_INFO;
-static tlog_format_func tlog_format;
-static unsigned int tlog_localtime_lock = 0;
+unsigned int tlog_localtime_lock;
 
 static const char *tlog_level_str[] = {
     "DEBUG",
@@ -175,7 +197,7 @@ static int _tlog_mkdir(const char *path)
         return 0;
     }
 
-    while (*path == ' ' && *path != '\0') {
+    while (*path == ' ') {
         path++;
     }
 
@@ -207,7 +229,6 @@ static int _tlog_mkdir(const char *path)
         }
 
         if (mkdir(path_c, 0750) != 0) {
-            fprintf(stderr, "create directory %s failed, %s\n", path_c, strerror(errno));
             return -1;
         }
 
@@ -306,10 +327,24 @@ void tlog_set_maxline_size(struct tlog_log *log, int size)
     log->max_line_size = size;
 }
 
-void tlog_set_permission(struct tlog_log *log, unsigned int file, unsigned int archive)
+void tlog_logcount(struct tlog_log *log, int count)
+{
+    if (log == NULL) {
+        return;
+    }
+
+    if (count < 0) {
+        count = 0;
+    }
+
+    log->logcount = count;
+}
+
+void tlog_set_permission(struct tlog_log *log, mode_t file, mode_t archive)
 {
     log->file_perm = file;
     log->archive_perm = archive;
+    log->mode_changed = 1;
 }
 
 int tlog_localtime(struct tlog_time *tm)
@@ -340,7 +375,7 @@ void *tlog_get_private(tlog_log *log)
     return log->private_data;
 }
 
-static int _tlog_format(char *buff, int maxlen, struct tlog_loginfo *info, void *userptr, const char *format, va_list ap)
+static int _tlog_root_default_format(char *buff, int maxlen, struct tlog_loginfo *info, void *userptr, const char *format, va_list ap)
 {
     int len = 0;
     int total_len = 0;
@@ -349,16 +384,18 @@ static int _tlog_format(char *buff, int maxlen, struct tlog_loginfo *info, void
 
     unused = userptr;
 
-    if (tlog.root->multi_log) {
-        /* format prefix */
-        len = snprintf(buff, maxlen, "[%.4d-%.2d-%.2d %.2d:%.2d:%.2d,%.3d][%5d][%4s][%17s:%-4d] ",
-            tm->year, tm->mon, tm->mday, tm->hour, tm->min, tm->sec, tm->usec / 1000, getpid(),
-            tlog_get_level_string(info->level), info->file, info->line);
-    } else {
-        /* format prefix */
-        len = snprintf(buff, maxlen, "[%.4d-%.2d-%.2d %.2d:%.2d:%.2d,%.3d][%5s][%17s:%-4d] ",
-            tm->year, tm->mon, tm->mday, tm->hour, tm->min, tm->sec, tm->usec / 1000,
-            tlog_get_level_string(info->level), info->file, info->line);
+    if (tlog.output_no_prefix == 0) {
+        if (tlog.root->multi_log) {
+            /* format prefix */
+            len = snprintf(buff, maxlen, "[%.4d-%.2d-%.2d %.2d:%.2d:%.2d,%.3d][%5d][%4s][%17s:%-4d] ",
+                tm->year, tm->mon, tm->mday, tm->hour, tm->min, tm->sec, tm->usec / 1000, getpid(),
+                tlog_get_level_string(info->level), info->file, info->line);
+        } else {
+            /* format prefix */
+            len = snprintf(buff, maxlen, "[%.4d-%.2d-%.2d %.2d:%.2d:%.2d,%.3d][%5s][%17s:%-4d] ",
+                tm->year, tm->mon, tm->mday, tm->hour, tm->min, tm->sec, tm->usec / 1000,
+                tlog_get_level_string(info->level), info->file, info->line);
+        }
     }
 
     if (len < 0 || len >= maxlen) {
@@ -388,7 +425,7 @@ static int _tlog_root_log_buffer(char *buff, int maxlen, void *userptr, const ch
     struct tlog_segment_log_head *log_head = NULL;
     int max_format_len = 0;
 
-    if (tlog_format == NULL) {
+    if (tlog.root_format == NULL) {
         return -1;
     }
 
@@ -400,7 +437,7 @@ static int _tlog_root_log_buffer(char *buff, int maxlen, void *userptr, const ch
 
     max_format_len = maxlen - len - 2;
     buff[maxlen - 1] = 0;
-    log_len = tlog_format(buff + len, max_format_len, &info_inter->info, info_inter->userptr, format, ap);
+    log_len = tlog.root_format(buff + len, max_format_len, &info_inter->info, info_inter->userptr, format, ap);
     if (log_len < 0) {
         return -1;
     } else if (log_len >= max_format_len) {
@@ -418,11 +455,10 @@ static int _tlog_root_log_buffer(char *buff, int maxlen, void *userptr, const ch
         log_len++;
     }
 
-    if (tlog.root->segment_log) {
+    if (tlog.root->segment_log && log_head != NULL) {
         if (len + 1 < maxlen - 1) {
             *(buff + len) = '\0';
             len++;
-            log_len++;
         }
         log_head->len = log_len;
     }
@@ -496,6 +532,10 @@ static int _tlog_vprintf(struct tlog_log *log, vprint_callback print_callback, v
         return -1;
     }
 
+    if (unlikely(log->logcount <= 0 && log->logscreen == 0)) {
+        return 0;
+    }
+
     if (_tlog_need_drop(log) == 0) {
         return -1;
     }
@@ -504,9 +544,12 @@ static int _tlog_vprintf(struct tlog_log *log, vprint_callback print_callback, v
     if (len <= 0) {
         return -1;
     } else if (len >= log->max_line_size) {
-        strncpy(buff, "[LOG TOO LONG, DISCARD]\n", sizeof(buff));
-        buff[sizeof(buff) - 1] = '\0';
-        len = strnlen(buff, sizeof(buff));
+        len = log->max_line_size;
+        buff[len - 1] = '\0';
+        buff[len - 2] = '\n';
+        buff[len - 3] = '.';
+        buff[len - 4] = '.';
+        buff[len - 5] = '.';
     }
 
     pthread_mutex_lock(&tlog.lock);
@@ -596,31 +639,100 @@ int tlog_printf(struct tlog_log *log, const char *format, ...)
     return len;
 }
 
-static int _tlog_early_print(const char *format, va_list ap)
+int tlog_stdout_with_color(tlog_level level, const char *buff, int bufflen)
+{
+    int unused __attribute__((unused));
+    const char *color = NULL;
+
+    switch (level) {
+    case TLOG_DEBUG:
+        color = "\033[0;94m";
+        break;
+    case TLOG_NOTICE:
+        color = "\033[0;97m";
+        break;
+    case TLOG_WARN:
+        color = "\033[0;33m";
+        break;
+    case TLOG_ERROR:
+        color = "\033[0;31m";
+        break;
+    case TLOG_FATAL:
+        color = "\033[31;1m";
+        break;
+    default:
+        unused = write(STDOUT_FILENO, buff, bufflen);
+        return bufflen;
+    }
+
+    if (color != NULL) {
+        fprintf(stdout, "%s%.*s\033[0m\n", color, bufflen - 1, buff);
+    } else {
+        fprintf(stdout, "%s", buff);
+    }
+
+    return bufflen;    
+}
+
+static int _tlog_early_print(struct tlog_info_inter *info_inter, const char *format, va_list ap)
 {
     char log_buf[TLOG_MAX_LINE_LEN];
     size_t len = 0;
     size_t out_len = 0;
+    struct tlog_time cur_time;
     int unused __attribute__((unused));
 
-    if (tlog_disable_early_print) {
+    if (tlog.early_print_disable) {
         return 0;
     }
 
-    len = vsnprintf(log_buf, sizeof(log_buf), format, ap);
+    if (_tlog_gettime(&cur_time) != 0) {
+        return -1;
+    }
+
+    if (tlog.tlog_early_print != NULL) {
+        tlog.tlog_early_print(&info_inter->info, format, ap);
+        return out_len;
+    }
+
+    if (tlog.early_print_no_prefix == 0) {
+        len = snprintf(log_buf, sizeof(log_buf), "[%.4d-%.2d-%.2d %.2d:%.2d:%.2d,%.3d][%5s][%17s:%-4d] ",
+            cur_time.year, cur_time.mon, cur_time.mday, cur_time.hour, cur_time.min, cur_time.sec, cur_time.usec / 1000,
+            tlog_get_level_string(info_inter->info.level), info_inter->info.file, info_inter->info.line);
+    }
+
     out_len = len;
+    len = vsnprintf(log_buf + out_len, sizeof(log_buf) - out_len - 1, format, ap);
+    out_len += len;
     if (len <= 0) {
         return -1;
-    } else if (len >= sizeof(log_buf)) {
-        out_len = sizeof(log_buf);
+    } else if (len >= sizeof(log_buf) - 1) {
+        out_len = sizeof(log_buf) - 1;
     }
 
-    unused = write(STDOUT_FILENO, log_buf, out_len);
     if (log_buf[out_len - 1] != '\n') {
-        unused = write(STDOUT_FILENO, "\n", 1);
+        log_buf[out_len] = '\n';
+        out_len++;
     }
 
-    return len;
+    if (out_len + 1 < sizeof(log_buf) - out_len - 1) {
+        log_buf[out_len] = '\0';
+    }
+
+    if (tlog.early_print_output != NULL) {
+        len = tlog.early_print_output(&info_inter->info, log_buf, out_len, tlog.early_print_userptr);
+        if (tlog.early_print_with_screen == 0) {
+            return len;
+        }
+    }
+
+    if (tlog.early_print_color) {
+        unused = tlog_stdout_with_color(info_inter->info.level, log_buf, out_len);
+    } else {
+        unused = write(STDOUT_FILENO, log_buf, out_len);
+    }
+
+    return out_len;
 }
 
 int tlog_vext(tlog_level level, const char *file, int line, const char *func, void *userptr, const char *format, va_list ap)
@@ -631,14 +743,6 @@ int tlog_vext(tlog_level level, const char *file, int line, const char *func, vo
         return 0;
     }
 
-    if (tlog.root == NULL) {
-        return _tlog_early_print(format, ap);
-    }
-
-    if (unlikely(tlog.root->logsize <= 0)) {
-        return 0;
-    }
-
     if (level >= TLOG_END) {
         return -1;
     }
@@ -652,6 +756,14 @@ int tlog_vext(tlog_level level, const char *file, int line, const char *func, vo
         return -1;
     }
 
+    if (tlog.root == NULL) {
+        return _tlog_early_print(&info_inter, format, ap);
+    }
+
+    if (unlikely(tlog.root->logsize <= 0)) {
+        return 0;
+    }
+
     return _tlog_vprintf(tlog.root, _tlog_root_log_buffer, &info_inter, format, ap);
 }
 
@@ -706,7 +818,7 @@ static int _tlog_list_dir(const char *path, list_callback callback, void *userpt
 
     dir = opendir(path);
     if (dir == NULL) {
-        fprintf(stderr, "open directory failed, %s\n", strerror(errno));
+        fprintf(stderr, "tlog: open directory failed, %s\n", strerror(errno));
         goto errout;
     }
 
@@ -779,7 +891,7 @@ static int _tlog_get_oldest_callback(const char *path, struct dirent *entry, voi
 
     if (oldestlog->mtime == 0 || oldestlog->mtime > sb.st_mtime) {
         oldestlog->mtime = sb.st_mtime;
-        strncpy(oldestlog->name, entry->d_name, sizeof(oldestlog->name) - 1);
+        strncpy(oldestlog->name, entry->d_name, sizeof(oldestlog->name));
         oldestlog->name[sizeof(oldestlog->name) - 1] = '\0';
         return 0;
     }
@@ -817,7 +929,7 @@ static int _tlog_remove_oldlog(struct tlog_log *log)
 
     /* get total log file number */
     if (_tlog_list_dir(log->logdir, _tlog_count_log_callback, &count_log) != 0) {
-        fprintf(stderr, "get log file count failed.\n");
+        fprintf(stderr, "tlog: get log file count failed.\n");
         return -1;
     }
 
@@ -854,7 +966,7 @@ static int _tlog_log_lock(struct tlog_log *log)
     snprintf(lock_file, sizeof(lock_file), "%s/%s.lock", log->logdir, log->logname);
     fd = open(lock_file, O_RDWR | O_CREAT | O_CLOEXEC, S_IRUSR | S_IWUSR);
     if (fd == -1) {
-        fprintf(stderr, "create pid file failed, %s", strerror(errno));
+        fprintf(stderr, "tlog: create lock file failed, %s", strerror(errno));
         return -1;
     }
 
@@ -889,14 +1001,14 @@ static void _tlog_wait_pid(struct tlog_log *log, int wait_hang)
     log->zip_pid = -1;
     char gzip_file[PATH_MAX * 2];
 
-    /* rename ziped file */
+    /* rename zipped file */
     snprintf(gzip_file, sizeof(gzip_file), "%s/%s.pending.gz", log->logdir, log->logname);
     if (_tlog_rename_logfile(log, gzip_file) != 0) {
         _tlog_log_unlock(log);
         return;
     }
 
-    /* remove oldes file */
+    /* remove oldest file */
     _tlog_remove_oldlog(log);
     _tlog_log_unlock(log);
 }
@@ -919,47 +1031,68 @@ static void _tlog_close_all_fd_by_res(void)
     }
 }
 
+static int _tlog_str_to_int(const char *str)
+{
+    int num = 0;
+
+    while (*str >= '0' && *str <= '9') {
+        num = num * 10 + (*str - '0');
+        ++str;
+    }
+
+    if (*str) {
+        return -1;
+    }
+
+    return num;
+}
+
 static void _tlog_close_all_fd(void)
 {
-    char path_name[PATH_MAX];
-    DIR *dir = NULL;
-    struct dirent *ent;
+#if defined(__linux__)
     int dir_fd = -1;
 
-    snprintf(path_name, sizeof(path_name), "/proc/self/fd/");
-    dir = opendir(path_name);
-    if (dir == NULL) {
+    dir_fd = open("/proc/self/fd/", O_RDONLY | O_DIRECTORY);
+    if (dir_fd < 0) {
         goto errout;
     }
 
-    dir_fd = dirfd(dir);
+    char buffer[sizeof(struct linux_dirent64)];
+    int bytes;
+    while ((bytes = syscall(SYS_getdents64, dir_fd,
+                (struct linux_dirent64 *)buffer,
+                sizeof(buffer)))
+        > 0) {
+        struct linux_dirent64 *entry;
+        int offset;
 
-    while ((ent = readdir(dir)) != NULL) {
-        int fd = atoi(ent->d_name);
-        if (fd < 0 || dir_fd == fd) {
-            continue;
-        }
-        switch (fd) {
-        case STDIN_FILENO:
-        case STDOUT_FILENO:
-        case STDERR_FILENO:
-            continue;
-            break;
-        default:
-            break;
-        }
+        for (offset = 0; offset < bytes; offset += entry->d_reclen) {
+            int fd;
+            entry = (struct linux_dirent64 *)(buffer + offset);
+            if ((fd = _tlog_str_to_int(entry->d_name)) < 0) {
+                continue;
+            }
 
-        close(fd);
+            if (fd == dir_fd || fd == STDIN_FILENO || fd == STDOUT_FILENO || fd == STDERR_FILENO) {
+                continue;
+            }
+            close(fd);
+        }
     }
 
-    closedir(dir);
 
+    if (bytes < 0) {
+        goto errout;
+    }
+
+    close(dir_fd);
     return;
 errout:
-    if (dir) {
-        closedir(dir);
+    if (dir_fd > 0) {
+        close(dir_fd);
+        dir_fd = -1;
     }
-
+#endif
     _tlog_close_all_fd_by_res();
     return;
 }
@@ -967,7 +1100,6 @@ errout:
 static int _tlog_archive_log_compressed(struct tlog_log *log)
 {
     char gzip_file[TLOG_BUFF_LEN];
-    char gzip_cmd[PATH_MAX * 2];
     char log_file[TLOG_BUFF_LEN];
     char pending_file[TLOG_BUFF_LEN];
 
@@ -995,14 +1127,19 @@ static int _tlog_archive_log_compressed(struct tlog_log *log)
     }
 
     /* start gzip process to compress log file */
-    snprintf(gzip_cmd, sizeof(gzip_cmd), "gzip -1 %s", pending_file);
     if (log->zip_pid <= 0) {
         int pid = vfork();
         if (pid == 0) {
             _tlog_close_all_fd();
-            execl("/bin/sh", "sh", "-c", gzip_cmd, NULL);
+            execl(tlog.gzip_cmd, tlog.gzip_cmd, "-1", pending_file, NULL);
+            fprintf(stderr, "tlog: execl gzip failed, no compress\n");
+            log->nocompress = 1;
             _exit(1);
         } else if (pid < 0) {
+            if (errno == EPERM || errno == EACCES) {
+                fprintf(stderr, "tlog: vfork failed, errno: %d, no compress\n", errno);
+                log->nocompress = 1;
+            }
             goto errout;
         }
         log->zip_pid = pid;
@@ -1039,7 +1176,7 @@ static int _tlog_archive_log_nocompress(struct tlog_log *log)
         goto errout;
     }
 
-    /* remove oldes file */
+    /* remove oldest file */
     _tlog_remove_oldlog(log);
     _tlog_log_unlock(log);
 
@@ -1059,7 +1196,7 @@ static int _tlog_archive_log(struct tlog_log *log)
     }
 }
 
-void _tlog_get_log_name_dir(struct tlog_log *log)
+static void _tlog_get_log_name_dir(struct tlog_log *log)
 {
     char log_file[PATH_MAX];
     if (log->fd > 0) {
@@ -1070,16 +1207,33 @@ void _tlog_get_log_name_dir(struct tlog_log *log)
     pthread_mutex_lock(&tlog.lock);
     strncpy(log_file, log->pending_logfile, sizeof(log_file) - 1);
     log_file[sizeof(log_file) - 1] = '\0';
-    strncpy(log->logdir, dirname(log_file), sizeof(log->logdir));
+    strncpy(log->logdir, dirname(log_file), sizeof(log->logdir) - 1);
     log->logdir[sizeof(log->logdir) - 1] = '\0';
     strncpy(log_file, log->pending_logfile, PATH_MAX);
     log_file[sizeof(log_file) - 1] = '\0';
-    strncpy(log->logname, basename(log_file), sizeof(log->logname));
+    strncpy(log->logname, basename(log_file), sizeof(log->logname) - 1);
     log->logname[sizeof(log->logname) - 1] = '\0';
     pthread_mutex_unlock(&tlog.lock);
 }
 
-static int _tlog_write(struct tlog_log *log, const char *buff, int bufflen)
+static int _tlog_write_screen(struct tlog_log *log, struct tlog_loginfo *info, const char *buff, int bufflen)
+{
+    if (bufflen <= 0) {
+        return 0;
+    }
+
+    if (log->logscreen == 0) {
+        return 0;
+    }
+
+    if (info == NULL) {
+        return write(STDOUT_FILENO, buff, bufflen);;
+    }
+
+    return tlog_stdout_with_color(info->level, buff, bufflen);
+}
+
+static int _tlog_write_ext(struct tlog_log *log, struct tlog_loginfo *info, const char *buff, int bufflen)
 {
     int len;
     int unused __attribute__((unused));
@@ -1093,9 +1247,8 @@ static int _tlog_write(struct tlog_log *log, const char *buff, int bufflen)
         log->rename_pending = 0;
     }
 
-    /* output log to screen */
-    if (log->logscreen) {
-        unused = write(STDOUT_FILENO, buff, bufflen);
+    if (log->logcount <= 0) {
+        return 0;
     }
 
     /* if log file size exceeds threshold, start to compress */
@@ -1117,7 +1270,6 @@ static int _tlog_write(struct tlog_log *log, const char *buff, int bufflen)
 
     if (log->fd <= 0) {
         /* open a new log file to write */
-        static int print_errmsg = 1;
         time_t now;
 
         time(&now);
@@ -1128,24 +1280,37 @@ static int _tlog_write(struct tlog_log *log, const char *buff, int bufflen)
 
         char logfile[PATH_MAX * 2];
         if (_tlog_mkdir(log->logdir) != 0) {
-            fprintf(stderr, "create log dir %s failed.\n", log->logdir);
+            if (log->print_errmsg == 0) {
+                return -1;
+            }
+            log->print_errmsg = 0;
+            fprintf(stderr, "tlog: create log dir %s failed, %s\n", log->logdir, strerror(errno));
+            if (errno == EACCES && log->logscreen == 0) {
+                fprintf(stderr, "tlog: no permission to write log file, output log to console\n");
+                tlog_logscreen(log, 1);
+                tlog_logcount(log, 0);
+            }
             return -1;
         }
         snprintf(logfile, sizeof(logfile), "%s/%s", log->logdir, log->logname);
         log->filesize = 0;
         log->fd = open(logfile, O_APPEND | O_CREAT | O_WRONLY | O_CLOEXEC, log->file_perm);
         if (log->fd < 0) {
-            if (print_errmsg == 0) {
+            if (log->print_errmsg == 0) {
                 return -1;
             }
 
-            fprintf(stderr, "open log file %s failed, %s\n", logfile, strerror(errno));
-            print_errmsg = 0;
+            fprintf(stderr, "tlog: open log file %s failed, %s\n", logfile, strerror(errno));
+            log->print_errmsg = 0;
             return -1;
         }
 
+        if (log->mode_changed != 0) {
+            fchmod(log->fd, log->file_perm);
+        }
+
         log->last_try = 0;
-        print_errmsg = 1;
+        log->print_errmsg = 1;
         /* get log file size */
         log->filesize = lseek(log->fd, 0, SEEK_END);
     }
@@ -1163,6 +1328,11 @@ static int _tlog_write(struct tlog_log *log, const char *buff, int bufflen)
     return len;
 }
 
+static inline int _tlog_write(struct tlog_log *log, const char *buff, int bufflen)
+{
+    return _tlog_write_ext(log, NULL, buff, bufflen);
+}
+
 int tlog_write(struct tlog_log *log, const char *buff, int bufflen)
 {
     return _tlog_write(log, buff, bufflen);
@@ -1296,12 +1466,26 @@ static struct tlog_log *_tlog_wait_log_locked(struct tlog_log *last_log)
     int ret = 0;
     struct timespec tm;
     struct tlog_log *log = NULL;
+    struct tlog_log *next = NULL;
+    int need_wait_pid = 0;
+
+    for (next = tlog.log; next != NULL; next = next->next) {
+        if (next->zip_pid > 0) {
+            need_wait_pid = 1;
+            break;
+        }
+    }
 
     clock_gettime(CLOCK_REALTIME, &tm);
     tm.tv_sec += 2;
     tlog.is_wait = 1;
     tlog.wait_on_log = last_log;
-    ret = pthread_cond_timedwait(&tlog.cond, &tlog.lock, &tm);
+    if (need_wait_pid != 0) {
+        ret = pthread_cond_timedwait(&tlog.cond, &tlog.lock, &tm);
+    } else {
+        ret = pthread_cond_wait(&tlog.cond, &tlog.lock);
+    }
+
     tlog.is_wait = 0;
     tlog.wait_on_log = NULL;
     errno = ret;
@@ -1323,6 +1507,19 @@ static void _tlog_wakeup_waiters(struct tlog_log *log)
     pthread_mutex_unlock(&log->lock);
 }
 
+static int _tlog_write_output_func(struct tlog_log *log, char *buff, int bufflen)
+{
+    if (log->logscreen && log != tlog.root) {
+        _tlog_write_screen(log, NULL, buff, bufflen);
+    }
+
+    if (log->output_func == NULL) {
+        return -1;
+    }
+
+    return log->output_func(log, buff, bufflen);
+}
+
 static void _tlog_write_one_segment_log(struct tlog_log *log, char *buff, int bufflen)
 {
     struct tlog_segment_head *segment_head = NULL;
@@ -1333,8 +1530,8 @@ static void _tlog_write_one_segment_log(struct tlog_log *log, char *buff, int bu
         if (segment_head->magic != TLOG_SEGMENT_MAGIC) {
             return;
         }
-
-        log->output_func(log, segment_head->data, segment_head->len - 1);
+        
+        _tlog_write_output_func(log, segment_head->data, segment_head->len - 1);
         write_len += segment_head->len + sizeof(*segment_head);
         segment_head = (struct tlog_segment_head *)(buff + write_len);
     }
@@ -1351,10 +1548,10 @@ static void _tlog_write_segments_log(struct tlog_log *log, int log_len, int log_
 
 static void _tlog_write_buff_log(struct tlog_log *log, int log_len, int log_extlen)
 {
-    log->output_func(log, log->buff + log->start, log_len);
+    _tlog_write_output_func(log, log->buff + log->start, log_len);
     if (log_extlen > 0) {
         /* write extend buffer log */
-        log->output_func(log, log->buff, log_extlen);
+        _tlog_write_output_func(log, log->buff, log_extlen);
     }
 }
 
@@ -1370,11 +1567,32 @@ static void _tlog_work_write(struct tlog_log *log, int log_len, int log_extlen,
     if (log_dropped > 0) {
         /* if there is dropped log, record dropped log number */
         char dropmsg[TLOG_TMP_LEN];
-        snprintf(dropmsg, sizeof(dropmsg), "[Totoal Dropped %d Messages]\n", log_dropped);
-        log->output_func(log, dropmsg, strnlen(dropmsg, sizeof(dropmsg)));
+        char *msg = dropmsg;
+        struct tlog_segment_log_head *log_head = NULL;
+        if (log->segment_log) {
+            memset(dropmsg, 0, sizeof(struct tlog_segment_log_head));
+            log_head = (struct tlog_segment_log_head *)dropmsg;
+            msg += sizeof(struct tlog_segment_log_head);
+            log_head->info.level = TLOG_WARN;
+        }
+
+        int len = snprintf(msg, msg - dropmsg, "[Total Dropped %d Messages]\n", log_dropped);
+        if (log_head) {
+            log_head->len = len;
+        }
+        _tlog_write_output_func(log, dropmsg, strnlen(dropmsg, sizeof(dropmsg)));
     }
 }
 
+static int _tlog_root_write_screen_log(struct tlog_log *log, struct tlog_loginfo *info, const char *buff, int bufflen)
+{
+    if (log->logscreen == 0) {
+        return 0;
+    }
+
+    return _tlog_write_screen(log, info, buff, bufflen);
+}
+
 static int _tlog_root_write_log(struct tlog_log *log, const char *buff, int bufflen)
 {
     struct tlog_segment_log_head *head = NULL;
@@ -1382,16 +1600,21 @@ static int _tlog_root_write_log(struct tlog_log *log, const char *buff, int buff
     if (tlog.output_func == NULL) {
         if (log->segment_log) {
             head = (struct tlog_segment_log_head *)buff;
-            return _tlog_write(log, head->data, head->len);
+            _tlog_root_write_screen_log(log, &head->info, head->data, head->len);
+            return _tlog_write_ext(log, &head->info, head->data, head->len);
         }
+        _tlog_root_write_screen_log(log, NULL, buff, bufflen);
         return _tlog_write(log, buff, bufflen);
     }
 
     if (log->segment_log && tlog.root == log) {
         head = (struct tlog_segment_log_head *)buff;
+        _tlog_root_write_screen_log(log, &head->info, head->data, head->len - 1);
         return tlog.output_func(&head->info, head->data, head->len - 1, tlog_get_private(log));
     }
 
+    _tlog_root_write_screen_log(log, NULL, buff, bufflen);
+    memset(&empty_info, 0, sizeof(empty_info));
     return tlog.output_func(&empty_info.info, buff, bufflen, tlog_get_private(log));
 }
 
@@ -1519,9 +1742,23 @@ static void *_tlog_work(void *arg)
     return NULL;
 }
 
-void tlog_set_early_printf(int enable)
+void tlog_set_early_printf(int enable, int no_prefix, int color)
 {
-    tlog_disable_early_print = (enable == 0) ? 1 : 0;
+    tlog.early_print_disable = (enable == 0) ? 1 : 0;
+    tlog.early_print_no_prefix = (no_prefix == 0) ? 0 : 1;
+    tlog.early_print_color = (color == 0 || isatty(STDOUT_FILENO) ==  0) ? 0 : 1;
+}
+
+void tlog_reg_early_printf_callback(tlog_early_print_func callback)
+{
+    tlog.tlog_early_print = callback;
+}
+
+void tlog_reg_early_printf_output_callback(tlog_log_output_func callback, int log_screen, void *private_data)
+{
+    tlog.early_print_output = callback;
+    tlog.early_print_userptr = private_data;
+    tlog.early_print_with_screen = (log_screen == 0) ? 0 : 1;
 }
 
 const char *tlog_get_level_string(tlog_level level)
@@ -1533,6 +1770,11 @@ const char *tlog_get_level_string(tlog_level level)
     return tlog_level_str[level];
 }
 
+void tlog_set_maxlog_count(int count)
+{
+    tlog_logcount(tlog.root, count);
+}
+
 static void _tlog_log_setlogscreen(struct tlog_log *log, int enable)
 {
     if (log == NULL) {
@@ -1565,7 +1807,7 @@ void tlog_logscreen(tlog_log *log, int enable)
     _tlog_log_setlogscreen(log, enable);
 }
 
-int tlog_reg_output_func(tlog_log *log, tlog_output_func output)
+static int _tlog_reg_output_func(tlog_log *log, tlog_output_func output)
 {
     if (log == NULL) {
         return -1;
@@ -1581,9 +1823,18 @@ int tlog_reg_output_func(tlog_log *log, tlog_output_func output)
     return 0;
 }
 
+int tlog_reg_output_func(tlog_log *log, tlog_output_func output)
+{
+    if (log == tlog.root) {
+        return -1;
+    }
+
+    return _tlog_reg_output_func(log, output);
+}
+
 int tlog_reg_format_func(tlog_format_func callback)
 {
-    tlog_format = callback;
+    tlog.root_format = callback;
     return 0;
 }
 
@@ -1604,6 +1855,15 @@ int tlog_setlevel(tlog_level level)
     return 0;
 }
 
+int tlog_log_enabled(tlog_level level)
+{
+    if (level >= TLOG_END) {
+        return 0;
+    }
+
+    return (tlog_set_level >= level) ? 1 : 0;
+}
+
 tlog_level tlog_getlevel(void)
 {
     return tlog_set_level;
@@ -1614,18 +1874,47 @@ void tlog_set_logfile(const char *logfile)
     tlog_rename_logfile(tlog.root, logfile);
 }
 
+static void _tlog_get_gzip_cmd_path(void)
+{
+    char *copy_path = NULL;
+    char gzip_cmd_path[PATH_MAX];
+    const char *env_path = getenv("PATH");
+    char *save_ptr = NULL;
+
+    if (env_path == NULL) {
+        env_path = "/bin:/usr/bin:/usr/local/bin";
+    }
+
+    copy_path = strdup(env_path);
+    if (copy_path == NULL) {
+        return;
+    }
+
+    for (char *tok = strtok_r(copy_path, ":", &save_ptr); tok; tok = strtok_r(NULL, ":", &save_ptr)) {
+        snprintf(gzip_cmd_path, sizeof(gzip_cmd_path), "%s/gzip", tok);
+        if (access(gzip_cmd_path, X_OK) != 0) {
+            continue;
+        }
+
+        snprintf(tlog.gzip_cmd, sizeof(tlog.gzip_cmd), "%s", gzip_cmd_path);
+        break;
+    }
+
+    free(copy_path);
+}
+
 tlog_log *tlog_open(const char *logfile, int maxlogsize, int maxlogcount, int buffsize, unsigned int flag)
 {
     struct tlog_log *log = NULL;
 
     if (tlog.run == 0) {
-        fprintf(stderr, "tlog is not initialized.");
+        fprintf(stderr, "tlog: tlog is not initialized.\n");
         return NULL;
     }
 
     log = (struct tlog_log *)malloc(sizeof(*log));
     if (log == NULL) {
-        fprintf(stderr, "malloc log failed.");
+        fprintf(stderr, "tlog: malloc log failed.\n");
         return NULL;
     }
 
@@ -1636,16 +1925,18 @@ tlog_log *tlog_open(const char *logfile, int maxlogsize, int maxlogcount, int bu
     log->dropped = 0;
     log->buffsize = (buffsize > 0) ? buffsize : TLOG_BUFF_SIZE;
     log->logsize = (maxlogsize >= 0) ? maxlogsize : TLOG_LOG_SIZE;
-    log->logcount = (maxlogcount > 0) ? maxlogcount : TLOG_LOG_COUNT;
+    log->logcount = (maxlogcount <= 0) ? 0 : maxlogcount;
     log->fd = -1;
     log->filesize = 0;
     log->zip_pid = -1;
     log->is_exit = 0;
     log->fail = 0;
+    log->print_errmsg = 1;
     log->waiters = 0;
     log->block = ((flag & TLOG_NONBLOCK) == 0) ? 1 : 0;
     log->nocompress = ((flag & TLOG_NOCOMPRESS) == 0) ? 0 : 1;
     log->logscreen = ((flag & TLOG_SCREEN) == 0) ? 0 : 1;
+    log->logscreen_color = ((flag & TLOG_SCREEN_COLOR) == 0 || isatty(STDOUT_FILENO) == 0) ? 0 : 1;
     log->multi_log = ((flag & TLOG_MULTI_WRITE) == 0) ? 0 : 1;
     log->segment_log = ((flag & TLOG_SEGMENT) == 0) ? 0 : 1;
     log->max_line_size = TLOG_MAX_LINE_LEN;
@@ -1653,6 +1944,15 @@ tlog_log *tlog_open(const char *logfile, int maxlogsize, int maxlogcount, int bu
     log->file_perm = S_IRUSR | S_IWUSR | S_IRGRP;
     log->archive_perm = S_IRUSR | S_IRGRP;
 
+    if (log->nocompress == 0 && tlog.gzip_cmd[0] == '\0') {
+        log->nocompress = 1;
+    }
+
+    if (log->logscreen_color == 1) {
+        log->logscreen = 1;
+        log->segment_log = 1;
+    }
+
     tlog_rename_logfile(log, logfile);
     if (log->nocompress) {
         strncpy(log->suffix, TLOG_SUFFIX_LOG, sizeof(log->suffix));
@@ -1662,7 +1962,7 @@ tlog_log *tlog_open(const char *logfile, int maxlogsize, int maxlogcount, int bu
 
     log->buff = (char *)malloc(log->buffsize);
     if (log->buff == NULL) {
-        fprintf(stderr, "malloc log buffer failed, %s\n", strerror(errno));
+        fprintf(stderr, "tlog: malloc log buffer failed, %s\n", strerror(errno));
         goto errout;
     }
 
@@ -1750,7 +2050,7 @@ static void tlog_fork_child(void)
     pthread_attr_init(&attr);
     int ret = pthread_create(&tlog.tid, &attr, _tlog_work, NULL);
     if (ret != 0) {
-        fprintf(stderr, "create tlog work thread failed, %s\n", strerror(errno));
+        fprintf(stderr, "tlog: create tlog work thread failed, %s\n", strerror(errno));
         goto errout;
     }
 
@@ -1771,21 +2071,20 @@ int tlog_init(const char *logfile, int maxlogsize, int maxlogcount, int buffsize
     int ret;
     struct tlog_log *log = NULL;
 
-    if (tlog_format != NULL) {
-        fprintf(stderr, "tlog already initilized.\n");
+    if (tlog.root_format != NULL) {
+        fprintf(stderr, "tlog: already initialized.\n");
         return -1;
     }
 
     if (buffsize > 0 && buffsize < TLOG_MAX_LINE_SIZE_SET * 2) {
-        fprintf(stderr, "buffer size is invalid.\n");
+        fprintf(stderr, "tlog: buffer size is invalid.\n");
         return -1;
     }
 
-    tlog_format = _tlog_format;
-
     memset(&tlog, 0, sizeof(tlog));
     tlog.is_wait = 0;
 
+    _tlog_get_gzip_cmd_path();
     pthread_attr_init(&attr);
     pthread_cond_init(&tlog.cond, NULL);
     pthread_mutex_init(&tlog.lock, NULL);
@@ -1793,32 +2092,42 @@ int tlog_init(const char *logfile, int maxlogsize, int maxlogcount, int buffsize
 
     log = tlog_open(logfile, maxlogsize, maxlogcount, buffsize, flag);
     if (log == NULL) {
-        fprintf(stderr, "init tlog root failed.\n");
+        fprintf(stderr, "tlog: init tlog root failed.\n");
         goto errout;
     }
-    tlog_reg_output_func(log, _tlog_root_write_log);
+    _tlog_reg_output_func(log, _tlog_root_write_log);
+
+    if ((flag & TLOG_NOCOMPRESS) == 0 && tlog.gzip_cmd[0] == '\0') {
+        fprintf(stderr, "tlog: can not find gzip command, disable compress.\n");
+    }
+
+    tlog.output_no_prefix = ((flag & TLOG_FORMAT_NO_PREFIX) == 0) ? 0 : 1;
+    tlog.root = log;
+    tlog.root_format = _tlog_root_default_format;
 
     ret = pthread_create(&tlog.tid, &attr, _tlog_work, NULL);
     if (ret != 0) {
-        fprintf(stderr, "create tlog work thread failed, %s\n", strerror(errno));
+        fprintf(stderr, "tlog: create tlog work thread failed, %s\n", strerror(errno));
         goto errout;
     }
 
-    tlog.root = log;
     if (flag & TLOG_SUPPORT_FORK) {
         pthread_atfork(&tlog_fork_prepare, &tlog_fork_parent, &tlog_fork_child);
     }
     return 0;
 errout:
-    if (tlog.tid > 0) {
+    if (tlog.tid) {
         void *retval = NULL;
         tlog.run = 0;
         pthread_join(tlog.tid, &retval);
+        tlog.tid = 0;
     }
 
     pthread_cond_destroy(&tlog.cond);
     pthread_mutex_destroy(&tlog.lock);
     tlog.run = 0;
+    tlog.root = NULL;
+    tlog.root_format = NULL;
 
     _tlog_close(log, 1);
 
@@ -1827,13 +2136,18 @@ errout:
 
 void tlog_exit(void)
 {
-    if (tlog.tid > 0) {
+    if (tlog.root_format == NULL) {
+        return;
+    }
+
+    if (tlog.tid) {
         void *ret = NULL;
         tlog.run = 0;
         pthread_mutex_lock(&tlog.lock);
         pthread_cond_signal(&tlog.cond);
         pthread_mutex_unlock(&tlog.lock);
         pthread_join(tlog.tid, &ret);
+        tlog.tid = 0;
     }
 
     tlog.root = NULL;
@@ -1843,4 +2157,7 @@ void tlog_exit(void)
 
     pthread_cond_destroy(&tlog.cond);
     pthread_mutex_destroy(&tlog.lock);
+
+    tlog.root_format = NULL;
+    tlog.is_wait = 0;
 }
